from functools import cache
from packaging.version import InvalidVersion
from packaging.version import Version

import requests


def is_valid_version(
    version: Version,
    min_version: Version | None = None,
    max_version: Version | None = None,
    allow_prerelease: bool = False,
) -> bool:
    """Check if version is valid."""
    status = True
    if version.is_prerelease:
        status = allow_prerelease
    if status and min_version:
        status = version >= min_version
    if status and max_version:
        status = version < max_version
    return status


def version_latest(
    versions: list[str],
    min_version: str | None = None,
    max_version: str | None = None,
    allow_prerelease: bool = False,
) -> str | None:
    min_version_ = Version(min_version) if min_version else None
    max_version_ = Version(max_version) if max_version else None
    versions_ = []
    for version in versions:
        try:
            v_info = (Version(version.replace("v", "")), version)
        except InvalidVersion:
            continue
        versions_.append(v_info)

    versions_ = sorted(versions_, reverse=True)
    valid = [
        (version, raw_version)
        for version, raw_version in versions_
        if is_valid_version(version, min_version_, max_version_, allow_prerelease)
    ]
    return valid[0][1] if valid else None


def get_pypi_package_versions(package: str) -> list[str]:
    """Get versions for a PyPi package."""
    url: str = f"https://pypi.org/pypi/{package}/json"
    resp = requests.get(url)  # noQA: S113
    data = resp.json()
    return list(data.get("releases").keys())


@cache
def package_versions(package_name: str = "Products.CMFPlone") -> list[str]:
    versions = sorted(get_pypi_package_versions(package_name))
    return versions


@cache
def plone_versions() -> list[str]:
    return package_versions("Products.CMFPlone")


def latest_package_version(package_name: str) -> str | None:
    """Latest version for base package."""
    versions = package_versions(package_name)
    return version_latest(versions, allow_prerelease=True)



from packaging.requirements import Requirement
from pathlib import Path
from repoplone import _types as t
from tomlkit import container
from tomlkit import items

import re
import requests
import tomlkit


def _process_requirements(data: tomlkit.TOMLDocument, key: str) -> list[Requirement]:
    requirements: list[Requirement] = []
    raw_dependencies: items.Table | items.Array = (
        _get_project_table(data).get(key.split(".")[-1])
        if key.startswith("project")
        else data.get(key)
    ) or tomlkit.table()
    if isinstance(raw_dependencies, items.Array):
        tmp_table = tomlkit.table()
        tmp_table.add(key, raw_dependencies)
        raw_dependencies = tmp_table
    if raw_dependencies:
        for group_values in raw_dependencies.values():
            for value in group_values:
                requirements.append(Requirement(value))
    return requirements


def _get_project_table(data: tomlkit.TOMLDocument) -> items.Table:
    """Return the current project information."""
    project = data["project"]
    if not isinstance(project, items.Table | container.OutOfOrderTableProxy):
        raise ValueError("Invalid data")
    return project


def _get_uv_table(data: tomlkit.TOMLDocument) -> items.Table:
    """Return the current project information."""
    tool: items.Table | None = data.get("tool")
    uv_config: items.Table | None = tool.get("uv") if tool else None
    if not uv_config:
        uv_config = tomlkit.table()
    return uv_config


def _get_project_dependencies(data: tomlkit.TOMLDocument) -> t.Requirements:
    """Return the current project dependencies."""
    raw_dependencies = _process_requirements(data, "project.dependencies")
    dependencies: t.Requirements = {req.name: req for req in raw_dependencies}
    return dependencies


def get_all_pinned_dependencies(data: tomlkit.TOMLDocument) -> t.Requirements:
    """Return dependencies that have a pinned version in pyproject.toml."""
    raw_dependencies: list[Requirement] = []
    for key in (
        "project.dependencies",
        "project.optional-dependencies",
        "dependency-groups",
    ):
        raw_dependencies.extend(_process_requirements(data, key))
    dependencies: t.Requirements = {
        req.name: req for req in raw_dependencies if req.specifier != ""
    }
    return dependencies


def get_remote_uv_dependencies(url: str) -> tuple[list[str], list[str]]:
    """Return dependencies listed under [tool.uv]."""
    response = requests.get(url)  # noQA: S113
    data = response.content.decode("utf-8")
    pyproject = _parse_pyproject(data)
    dependencies = [
        str(req)
        for req in _process_requirements(pyproject, "project.dependencies")
        if req.specifier != ""
    ]
    uv_config = _get_uv_table(pyproject)
    constraints = list(uv_config.get("constraint-dependencies") or tomlkit.array())
    return dependencies, constraints


def current_base_package(pyproject: Path, package_name: str) -> str | None:
    """Return the current base package version."""
    data = tomlkit.parse(pyproject.read_text())
    deps: t.Requirements = _get_project_dependencies(data)
    req = deps.get(package_name)
    if req:
        return str(req.specifier)[2:]
    return None


def _update_dependency(data: tomlkit.TOMLDocument, package: str, version: str) -> None:
    project = _get_project_table(data)
    deps = tomlkit.array()
    deps.multiline(True)
    for dep in _get_project_dependencies(data):
        if re.match(f"^{package}", dep):
            dep = f"{package}=={version}"
        deps.append(dep)
    project.update({"dependencies": deps})


def _update_constraints(data: tomlkit.TOMLDocument, raw_constraints: list[str]) -> None:
    tool_uv = data.get("tool", {}).get("uv", {})
    if not tool_uv:
        tool_uv = tomlkit.table(False)
        data.append("tool.uv", tool_uv)
    constraints = tomlkit.array()
    constraints.multiline(True)
    for line in raw_constraints:
        item = tomlkit.item(line)
        constraints.append(item)
    tool_uv.update({"constraint-dependencies": constraints})


def _parse_pyproject(src: str) -> tomlkit.TOMLDocument:
    return tomlkit.parse(src)


def parse_pyproject(pyproject: Path) -> tomlkit.TOMLDocument:
    return _parse_pyproject(pyproject.read_text())


def update_pyproject(
    pyproject: Path, package_name: str, version: str, constraints: list[str]
):
    """Update pyproject.toml with a new version of the package."""
    data: tomlkit.TOMLDocument = tomlkit.parse(pyproject.read_text())
    # Update dependency
    _update_dependency(data, package_name, version)
    # Constraints
    _update_constraints(data, constraints)
    # Update pyproject
    pyproject.write_text(tomlkit.dumps(data))



from ._git import repo_for_project
from ._git import repo_has_version
from ._hatch import get_hatch
from ._path import change_cwd
from .changelog import update_backend_changelog
from .versions import convert_python_node_version
from .versions import update_backend_version
from repoplone import _types as t
from repoplone import logger

import subprocess


def release_backend(settings: t.RepositorySettings, version: str, dry_run: bool):
    package = settings.backend
    # Update backend version
    hatch = get_hatch()
    if not dry_run:
        update_backend_version(package.path, version)
        update_backend_changelog(settings, dry_run, version)
    if not package.publish:
        return
    with change_cwd(settings.backend.path):
        logger.info(f"Build backend package {settings.backend.name}")
        # Build package
        hatch("build")
        if not dry_run:
            logger.info(f"Publish backend package {settings.backend.name}")
            hatch("publish")


def release_frontend(
    settings: t.RepositorySettings, project_version: str, dry_run: bool
):
    version = convert_python_node_version(project_version)
    action = "dry-release" if dry_run else "release"
    package = settings.frontend
    volto_addon_name = package.name
    logger.debug(f"Frontend: {action} for package {volto_addon_name} ({version})")
    cmd = "npx release-it --ci --no-git --no-github.release"
    if dry_run:
        # No need to check if we are authenticated on NPM
        cmd += " --dry-run --npm.skipChecks"
    elif not package.publish:
        cmd += "  --npm.skipChecks --no-npm.publish"
    cmd += f" -i {version}"
    result = subprocess.run(  # noQA: S602
        cmd,
        capture_output=True,
        text=True,
        shell=True,
        cwd=settings.frontend.path,
    )
    if result.returncode:
        raise RuntimeError(f"Frontend release failed {result.stderr}")


def valid_next_version(settings: t.RepositorySettings, next_version: str) -> bool:
    """Check if next version is valid."""
    is_valid = True
    repo = repo_for_project(settings.root_path)
    if repo:
        is_valid = not (repo_has_version(repo, next_version))
    return is_valid



from ._path import change_cwd
from click.testing import CliRunner
from datetime import datetime
from pathlib import Path
from repoplone import _types as t
from repoplone import utils
from towncrier._builder import find_fragments
from towncrier._settings import load_config_from_options
from towncrier._shell import cli as towncrier_app


CHANGELOG_PLACEHOLDER = "<!-- towncrier release notes start -->"


def _cleanup_draft(text: str, include_version: bool = False) -> str:
    lines = text.split("\n")
    for idx, line in enumerate(lines):
        if line.startswith("## "):
            idx += 0 if include_version else 1
            break
    text = "\n".join(lines[idx:])
    return text


def _prepare_section_changelog(text: str) -> str:
    """Prepare section changelog to be added to the project changelog."""
    text = _cleanup_draft(text)
    # Increase header levels
    text = text.replace("###", "####")
    if not text.strip():
        text = "\nNo significant changes.\n\n"
    return text


def _run_towncrier(config: Path, name: str, version: str, draft: bool = True) -> str:
    cwd = config.parent
    runner = CliRunner()
    args = [
        "build",
        "--config",
        f"{config}",
        "--yes",
        "--name",
        f"'{name}'",
        "--version",
        version,
    ]
    if draft:
        args.append("--draft")
    with change_cwd(cwd):
        result = runner.invoke(towncrier_app, args)
    return result.stdout


def generate_section_changelogs(
    settings: t.RepositorySettings, version: str = ""
) -> dict[str, dict]:
    sections = {}
    config: t.TowncrierSettings = settings.towncrier
    for section in config.sections:
        config_path = section.path
        result = _run_towncrier(
            config_path, name=settings.name, version=version, draft=True
        )
        sections[section.section_id] = {
            "name": section.name,
            "changes": _prepare_section_changelog(result),
        }
    return sections


def _find_fragments(path: Path, towncrier_settings: Path) -> list[tuple[str, str]]:
    base_directory, config = load_config_from_options(path, towncrier_settings)
    _, fragments = find_fragments(base_directory, config, False)
    return fragments


def _cleanup_news(path: Path, towncrier_settings: Path):
    all_fragments = _find_fragments(path, towncrier_settings)
    for raw_fragment_path, _ in all_fragments:
        fragment_path = Path(raw_fragment_path).resolve()
        # Remove
        fragment_path.unlink()


# Update Changelog at root
def _update_project_changelog(
    settings: t.RepositorySettings,
    sections: dict[str, dict],
    draft: bool = True,
    version: str = "",
) -> tuple[str, str]:
    root_changelog = settings.changelogs.root
    changelog_text = root_changelog.read_text()
    header = f"## {version} ({datetime.now():%Y-%m-%d})"
    new_entry = f"{header}\n"
    has_root = False
    for section_id, section_data in sections.items():
        if section_id == "repository":
            has_root = True
        section_name = section_data["name"]
        text = section_data["changes"]
        new_entry = f"{new_entry}\n### {section_name}\n{text}"

    text = f"{changelog_text}".replace(
        CHANGELOG_PLACEHOLDER, f"{CHANGELOG_PLACEHOLDER}\n{new_entry}"
    )
    if not draft:
        root_changelog.write_text(text)
        if has_root:
            # Cleanup top-level news folder
            _cleanup_news(settings.root_path, settings.towncrier.repository.path)

    return new_entry, text


def update_backend_changelog(
    settings: t.RepositorySettings, draft: bool = True, version: str = ""
) -> str:
    config_path = settings.towncrier.backend.path
    result = _run_towncrier(
        config_path, name=settings.name, version=version, draft=draft
    )
    if draft:
        result = _cleanup_draft(result, True)
    return result


def update_changelog(
    settings: t.RepositorySettings, draft: bool = True, version: str = ""
) -> tuple[str, str]:
    if draft and not version:
        version = utils.get_next_version(settings)
    sections = generate_section_changelogs(settings=settings, version=version)
    return _update_project_changelog(
        settings=settings, sections=sections, version=version, draft=draft
    )



from repoplone import _types as t

import os
import re
import requests


def get_token() -> str | None:
    """Check if this environment has the GITHUB_TOKEN set."""
    return os.getenv("GITHUB_TOKEN")


def gh_session() -> requests.Session | None:
    """Return a Session with correct headers for a GitHub REST API call."""
    session = None
    if token := get_token():
        session = requests.Session()
        session.headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {token}",
            "X-GitHub-Api-Version": "2022-11-28",
        }
    return session


def _get_owner_repo(remote_origin: str) -> str:
    """Return OWNER/REPO from a remote_origin."""
    pattern = r"^(git@|https:\/\/)github\.com(:|\/)(?P<owner>[^\/]*)\/(?P<repo>.+)$"
    match = re.match(pattern, remote_origin)
    if not match:
        raise ValueError(f"Not a valid GitHub repository {remote_origin}")
    groups = match.groupdict()
    owner = groups["owner"]
    repo = groups["repo"]
    if repo.endswith(".git"):
        repo = repo[:-4]
    return f"{owner}/{repo}"


def check_token(settings: t.RepositorySettings) -> bool:
    """Check if this environment has the GITHUB_TOKEN set."""
    token = get_token()
    status = False
    if token and (session := gh_session()):
        remote_path = _get_owner_repo(settings.remote_origin)
        response = session.get(f"https://api.github.com/repos/{remote_path}/releases")
        status = response.status_code == 200
    return status


def create_release(
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
) -> str:
    if not check_token(settings):
        return (
            f"Release {next_version} not created because GITHUB_TOKEN"
            "is not present or does not have correct permissions"
        )
    changelog = settings._tmp_changelog
    remote_path = _get_owner_repo(settings.remote_origin)
    if session := gh_session():
        payload = {
            "tag_name": next_version,
            "target_commitish": "main",
            "name": next_version,
            "body": changelog,
            "draft": False,
            "prerelease": False,
            "generate_release_notes": False,
        }
        response = session.post(
            f"https://api.github.com/repos/{remote_path}/releases", json=payload
        )
        if response.status_code == 201:
            data = response.json()
            url = data.get("html_url")
            msg = f"Release {next_version} created at {url}"
        else:
            data = response.json()
            msg = (
                f"Release {next_version} failed to be created "
                f"({response.status_code} {data['message']})"
            )
    else:
        msg = f"Release {next_version} failed to be created."
    # Remove tmp changelog
    settings._tmp_changelog = ""
    return msg



from dynaconf import Dynaconf


def parse_config() -> Dynaconf:
    """Parse repo settings."""
    settings = Dynaconf(
        settings_files=["repository.toml"],
        merge_enabled=False,
    )
    return settings



from .parser import parse_config
from repoplone import _types as t
from repoplone import utils
from repoplone.utils import _git as git_utils
from repoplone.utils._path import get_root_path


def get_settings() -> t.RepositorySettings:
    """Return base settings."""
    root_path = get_root_path()
    raw_settings = parse_config()
    try:
        name = raw_settings.repository.name
    except AttributeError:
        raise RuntimeError() from None
    managed_by_uv = bool(raw_settings.repository.get("managed_by_uv", False))
    root_changelog = root_path / raw_settings.repository.changelog
    version_path = root_path / raw_settings.repository.version
    version = version_path.read_text().strip()
    version_format = raw_settings.repository.get("version_format", "semver")
    compose_path = root_path / raw_settings.repository.compose
    repository_towncrier = raw_settings.repository.get("towncrier", {})
    backend = utils.get_backend(root_path, raw_settings)
    frontend = utils.get_frontend(root_path, raw_settings)
    towncrier = utils.get_towncrier_settings(backend, frontend, repository_towncrier)
    changelogs = utils.get_changelogs(root_changelog, backend, frontend)
    remote_origin = git_utils.remote_origin(root_path)
    return t.RepositorySettings(
        name=name,
        managed_by_uv=managed_by_uv,
        root_path=root_path,
        version=version,
        version_format=version_format,
        backend=backend,
        frontend=frontend,
        version_path=version_path,
        compose_path=compose_path,
        towncrier=towncrier,
        changelogs=changelogs,
        remote_origin=remote_origin,
    )



from ._hatch import get_hatch
from ._path import change_cwd
from hatchling.version.scheme import standard
from packaging.version import Version as PyPIVersion
from pathlib import Path
from repoplone import _types as t

import json
import re
import semver


VERSION_PATTERNS = (
    (r"^(a)(\d{1,2})", r"alpha.\2"),
    (r"^(b)(\d{1,2})", r"beta.\2"),
    (r"^(rc)(\d{1,2})", r"rc.\2"),
)

BUMPS = [
    "release",
    "major",
    "minor",
    "micro",
    "patch",
    "fix",
    "a",
    "b",
    "rc",
    "post",
    "dev",
]


def convert_python_node_version(version: str) -> str:
    """Converts a PyPI version into a semver version

    :param ver: the PyPI version
    :return: a semver version
    :raises ValueError: if epoch or post parts are used
    """
    pypi_version = PyPIVersion(version)
    pre = None if not pypi_version.pre else "".join([str(i) for i in pypi_version.pre])
    if pre:
        for raw_pattern, replace in VERSION_PATTERNS:
            pattern = re.compile(raw_pattern)
            if re.search(pattern, pre):
                pre = re.sub(pattern, replace, pre)

    parts = list(pypi_version.release)
    if len(parts) == 2:
        parts.append(0)
    major, minor, patch = parts
    version = str(
        semver.Version(major, minor, patch, prerelease=pre, build=pypi_version.dev)
    )
    return version


def get_repository_version(settings: t.RepositorySettings) -> str:
    """Return the currect repository version."""
    version_path = settings.version_path
    return version_path.read_text().strip()


def get_backend_version(backend_path: Path) -> str:
    """Get the current version used by the backend."""
    hatch = get_hatch()
    with change_cwd(backend_path):
        result = hatch("version")
    return result.stdout.strip()


def get_frontend_version(frontend_package_path: Path) -> str:
    """Get the current version used by the frontend."""
    package_json = (frontend_package_path / "package.json").resolve()
    package_data = json.loads(package_json.read_text())
    return package_data["version"]


def update_backend_version(backend_path: Path, version: str) -> str:
    """Update version used by the backend."""
    hatch = get_hatch()
    with change_cwd(backend_path):
        result = hatch("version", version)
    if result.exit_code:
        raise RuntimeError("Error setting backend version")
    return get_backend_version(backend_path)


def next_version(desired_version: str, original_version: str) -> str:
    """Return the next version for this project.

    desired_version could be either a full version or one of the
    version segments detailed here: https://hatch.pypa.io/1.12/version/#updating
    """
    scheme = standard.StandardScheme("", {})
    next_version = scheme.update(desired_version, original_version, {})
    return next_version


def report_cur_versions(settings: t.RepositorySettings) -> dict:
    sections: list[dict] = []
    cur_versions = {
        "repository": {"title": "Repository", "version": settings.version},
        "sections": sections,
    }
    for title, section in (
        ("Repository", settings),
        ("Backend", settings.backend),
        ("Frontend", settings.frontend),
    ):
        sections.append({
            "title": title,
            "name": section.name,
            "version": section.version,
        })
    return cur_versions


def report_next_versions(settings: t.RepositorySettings):
    cur_version = settings.version
    versions = []
    for bump in BUMPS:
        nv = next_version(bump, cur_version)
        nv_semver = convert_python_node_version(nv)
        versions.append({
            "bump": bump,
            "repository": nv,
            "backend": nv,
            "frontend": nv_semver,
        })
    return versions



from .pyproject import get_remote_uv_dependencies
from .versions import package_versions
from mxdev.processing import resolve_dependencies
from packaging.requirements import Requirement
from repoplone import _types as t


PACKAGE_CONSTRAINTS = {
    "plone": {
        "type": "pip",
        "url": "https://dist.plone.org/release/{version}/constraints.txt",
    },
    "Products.CMFPlone": {
        "type": "pip",
        "url": "https://dist.plone.org/release/{version}/constraints.txt",
    },
    "kitconcept.intranet": {
        "type": "uv",
        "url": "https://raw.githubusercontent.com/kitconcept/kitconcept.intranet/refs/tags/{version}/backend/pyproject.toml",
    },
    "portalbrasil.core": {
        "type": "uv",
        "url": "https://raw.githubusercontent.com/portal-br/core/refs/tags/{version}/backend/pyproject.toml",
    },
}


def _get_uv_constraints(url: str, package_name: str, version: str) -> list[str]:
    """Parse constraints inside a remote pyproject.toml file."""
    dependencies, constraints = get_remote_uv_dependencies(url)
    constraints.append(f"{package_name}=={version}")
    return parse_constraints(constraints, dependencies)


def _process_constraint(src: str) -> tuple[str, str]:
    req = Requirement(src)
    return req.name, str(req)


def parse_constraints(lines: list[str], existing: list[str]) -> list[str]:
    constraints = []
    existing_ = dict([_process_constraint(line) for line in existing])
    for line in lines:
        line = line.strip()
        if line.startswith("#") or not line:
            continue
        req_name, _ = _process_constraint(line)
        constraints.append(existing_.pop(req_name, line))
    if existing_:
        constraints.extend(existing_.values())
    return sorted(constraints, key=lambda x: x.lower())


def get_base_constraints(package_name: str, version: str) -> list[str]:
    pkg_config = PACKAGE_CONSTRAINTS.get(package_name)
    if not pkg_config:
        raise AttributeError(f"{package_name} is not supported at the moment.")
    constraints_type = pkg_config["type"]
    constraints_url = pkg_config["url"].format(version=version)
    match constraints_type:
        case "pip":
            _, constraints = resolve_dependencies(constraints_url, [], [], [], "c")
        case "uv":
            constraints = _get_uv_constraints(constraints_url, package_name, version)
        case _:
            raise AttributeError(
                f"{package_name} has an invalid constraints type: {constraints_type}."
            )
    return constraints


def get_package_constraints(
    package_name: str, version: str, existing_pins: t.Requirements
) -> list[str]:
    """Return plone constraints for a version."""
    versions = package_versions(package_name)
    existing_constraints = [
        str(v) for k, v in existing_pins.items() if k != package_name
    ]
    if version not in versions:
        raise RuntimeError(f"{package_name} {version} not available.")

    constraints = get_base_constraints(package_name, version)
    return parse_constraints(constraints, existing_constraints)



from .constraints import get_package_constraints
from .pyproject import current_base_package
from .pyproject import get_all_pinned_dependencies
from .pyproject import parse_pyproject
from .pyproject import update_pyproject
from .versions import latest_package_version


__all__ = [
    "current_base_package",
    "get_all_pinned_dependencies",
    "get_package_constraints",
    "latest_package_version",
    "parse_pyproject",
    "update_pyproject",
]



from collections.abc import Callable
from dynaconf import Dynaconf
from dynaconf.utils.boxing import DynaBox
from pathlib import Path
from repoplone import _types as t
from repoplone.utils import versions


PYPROJECT_TOML = "pyproject.toml"


def get_changelogs(
    root_changelog: Path, backend: t.Package, frontend: t.Package
) -> t.Changelogs:
    return t.Changelogs(root_changelog, backend.changelog, frontend.changelog)


def get_towncrier_settings(
    backend: t.Package, frontend: t.Package, repository: dict
) -> t.TowncrierSettings:
    sections = []
    raw_sections = [
        ("backend", "Backend", backend.towncrier),
        ("frontend", "Frontend", frontend.towncrier),
    ]
    if repository:
        section = (
            "repository",
            repository["section"],
            Path(repository["settings"]).resolve(),
        )
        raw_sections.append(section)
    sections = [t.TowncrierSection(*info) for info in raw_sections]
    return t.TowncrierSettings(sections=sections)


def get_pyproject(settings: t.RepositorySettings) -> Path | None:
    """Return the pyproject.toml for a monorepo."""
    paths = [
        settings.root_path,
        settings.backend.path,
    ]
    for base_path in paths:
        path = base_path / PYPROJECT_TOML
        if path.exists():
            return path
    return None


def get_next_version(settings: t.RepositorySettings) -> str:
    version_file = settings.version_path
    cur_version = version_file.read_text().strip()
    next_version = cur_version.replace(".dev", "")
    return next_version


def _get_package_info(
    root_path: Path,
    package_settings: DynaBox,
    default_base_package: str,
    version_func: Callable,
) -> t.Package:
    """Return package information for the frontend."""
    path = (root_path / package_settings.path).resolve()
    changelog = (root_path / package_settings.changelog).resolve()
    towncrier = (root_path / package_settings.towncrier_settings).resolve()
    version = version_func(path)
    publish = bool(package_settings.get("publish", True))
    base_package = package_settings.get("base_package", default_base_package)
    return t.Package(
        name=package_settings.name,
        path=path,
        base_package=base_package,
        version=version,
        publish=publish,
        changelog=changelog,
        towncrier=towncrier,
    )


def get_backend(root_path: Path, raw_settings: Dynaconf) -> t.Package:
    """Return package information for the backend."""
    package_settings = raw_settings.backend.package
    version_func = versions.get_backend_version
    default_base_package: str = "Products.CMFPlone"
    return _get_package_info(
        root_path, package_settings, default_base_package, version_func
    )


def get_frontend(root_path: Path, raw_settings: Dynaconf) -> t.Package:
    """Return package information for the frontend."""
    package_settings = raw_settings.frontend.package
    version_func = versions.get_frontend_version
    default_base_package: str = "@plone/volto"
    return _get_package_info(
        root_path, package_settings, default_base_package, version_func
    )



from dataclasses import dataclass
from packaging.requirements import Requirement
from pathlib import Path


Requirements = dict[str, Requirement]


@dataclass
class Changelogs:
    """Changelog locations."""

    root: Path
    backend: Path
    frontend: Path

    def sanity(self) -> bool:
        return self.root.exists() and self.backend.exists() and self.frontend.exists()


@dataclass
class Package:
    """Package information."""

    name: str
    path: Path
    changelog: Path
    towncrier: Path
    base_package: str
    publish: bool = True
    version: str = ""

    def sanity(self) -> bool:
        return (
            self.path.exists() and self.changelog.exists() and self.towncrier.exists()
        )


@dataclass
class TowncrierSection:
    """Towncrier section."""

    section_id: str
    name: str
    path: Path

    def sanity(self) -> bool:
        return self.path.exists() if self.path else False


@dataclass
class TowncrierSettings:
    """Towncrier settings."""

    sections: list[TowncrierSection]

    def __getattr__(self, name: str):
        for section in self.sections:
            if section.section_id == name:
                return section
        raise AttributeError(f"{name} not found")

    def sanity(self) -> bool:
        sections = self.sections
        checks = [section.sanity() for section in sections]
        return all(checks)


@dataclass
class RepositorySettings:
    """Settings for a distribution."""

    name: str
    managed_by_uv: bool
    root_path: Path
    version: str
    version_format: str
    backend: Package
    frontend: Package
    version_path: Path
    compose_path: Path
    towncrier: TowncrierSettings
    changelogs: Changelogs
    remote_origin: str
    _tmp_changelog: str = ""

    @property
    def path(self) -> Path:
        return self.root_path

    def sanity(self) -> bool:
        steps = [
            self.root_path.exists(),
            self.backend.sanity(),
            self.frontend.sanity(),
            self.version_path.exists(),
            self.compose_path.exists(),
            self.towncrier.sanity(),
            self.changelogs.sanity(),
        ]
        return all(steps)


@dataclass
class CTLContextObject:
    """Context object used by cli."""

    settings: RepositorySettings



from contextlib import contextmanager
from functools import cache
from pathlib import Path

import os


@contextmanager
def change_cwd(path: Path):
    """Sets the cwd within the context."""
    origin = Path().cwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


@cache
def get_root_path() -> Path:
    return (Path().cwd()).resolve()



from repoplone import _types as t
from repoplone.utils import display as dutils
from repoplone.utils import versions as vutils

import typer


app = typer.Typer()


@app.command()
def current(ctx: typer.Context):
    """Report versions of all components of this repository."""
    settings: t.RepositorySettings = ctx.obj.settings
    cur_versions = vutils.report_cur_versions(settings)
    title = "Current versions"
    cols = [{"header": "Section"}, {"header": "Name"}, {"header": "Version"}]
    rows = [
        (section["title"], section["name"], section["version"])
        for section in cur_versions["sections"]
    ]
    table = dutils.table(title, cols, rows)
    dutils.print(table)


@app.command(name="next")
def next_version(ctx: typer.Context):
    """Report versions of all components of this repository."""
    settings: t.RepositorySettings = ctx.obj.settings
    versions = vutils.report_next_versions(settings)
    title = "Possible next version"
    cols = [
        {"header": "Current Version"},
        {"header": "Desidered Version"},
        {"header": "Repository"},
        {"header": "Backend"},
        {"header": "Frontend"},
    ]
    rows = [
        (
            settings.version,
            version["bump"],
            version["repository"],
            version["backend"],
            version["frontend"],
        )
        for version in versions
    ]
    table = dutils.table(title, cols, rows)
    dutils.print(table)



from git import InvalidGitRepositoryError
from git import Remote
from git import Repo
from git import Tag
from pathlib import Path
from repoplone import logger


def _initialize_repo_for_project(path: Path) -> Repo:
    """Return the repository for a project."""
    repo = Repo.init(path)
    return repo


def repo_for_project(path: Path) -> Repo:
    """Return the repository for a project."""
    repo = Repo(path)
    return repo


def _get_remote(repo: Repo) -> Remote | None:
    try:
        origin = repo.remote("origin")
    except ValueError:
        logger.info("No origin for this repo")
        origin = None
    return origin


def remote_origin(path: Path) -> str:
    """Return the url for the remote origin."""
    try:
        repo = repo_for_project(path)
        origin = _get_remote(repo)
    except InvalidGitRepositoryError:
        origin = None
    return origin.url if origin else ""


def push_changes(repo: Repo, ref: Tag | None = None):
    if not (origin := _get_remote(repo)):
        return
    if ref:
        origin.push(ref)
    else:
        origin.push()


def commit_pending_changes(repo: Repo, message: str):
    git_cmd = repo.git
    git_cmd.commit("-am", message)


def repo_has_version(repo: Repo, version: str) -> bool:
    # Fetch existing tags
    origin = repo.remote("origin")
    if origin:
        origin.fetch()
    # List tags
    tags: list[Tag] = repo.tags
    names = [tag.name for tag in tags]
    return bool(version in names)


def create_version_tag(repo: Repo, version: str, message: str) -> Tag:
    # Create tag
    tag = repo.create_tag(version, message=message)
    # Push tag
    push_changes(repo, tag)
    return tag


def finish_release(repo: Repo, version: str) -> Tag:
    message = f"Release {version}"
    commit_pending_changes(repo, message)
    tag = create_version_tag(repo, version, message)
    push_changes(repo)
    return tag



from repoplone import __version__
from repoplone import _types as t
from repoplone.commands.dependencies import app as app_deps
from repoplone.commands.release import app as app_release
from repoplone.commands.versions import app as app_versions
from repoplone.settings import get_settings
from typing import Annotated

import typer


app = typer.Typer(no_args_is_help=True)


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: Annotated[
        bool, typer.Option(help="Report the version of this app.")
    ] = False,
):
    """Welcome to Plone Distribution Helper."""
    try:
        settings = get_settings()
    except RuntimeError:
        typer.echo("Not running inside a mono repo.")
        raise typer.Exit() from None
    if version:
        typer.echo(f"repoplone {__version__}")
    else:
        ctx_obj = t.CTLContextObject(settings=settings)
        ctx.obj = ctx_obj
        ctx.ensure_object(t.CTLContextObject)


app.add_typer(
    app_release, name="release", no_args_is_help=True, help="Release mono repo packages"
)
app.add_typer(
    app_deps,
    name="deps",
    no_args_is_help=True,
    help="Check and manage dependencies",
)
app.add_typer(
    app_versions,
    name="versions",
    no_args_is_help=True,
    help="Display version information about this repository",
)


def cli():
    app()


__all__ = ["cli"]



from repoplone import _types as t
from repoplone import logger
from repoplone import utils
from repoplone.utils import dependencies
from typing import Annotated

import typer


app = typer.Typer()


@app.command()
def info(ctx: typer.Context):
    """Report the base package in use."""
    settings: t.RepositorySettings = ctx.obj.settings
    if not settings.managed_by_uv:
        typer.echo("Only available for installations managed by uv.")
        raise typer.Exit(1)
    package_name: str = settings.backend.base_package
    logger.info(f"The base package is {package_name}")


@app.command()
def check(ctx: typer.Context):
    """Check latest version of base package and compare it to our current pinning."""
    settings: t.RepositorySettings = ctx.obj.settings
    if not settings.managed_by_uv:
        typer.echo("Only available for installations managed by uv.")
        raise typer.Exit(1)
    package_name: str = settings.backend.base_package
    pyproject = utils.get_pyproject(settings)
    current = (
        dependencies.current_base_package(pyproject, package_name)
        if pyproject
        else None
    )
    if not current:
        logger.info(f"{package_name} is not present in {pyproject}")
    else:
        latest_version = dependencies.latest_package_version(package_name)
        logger.info(f"Current version {current}, latest version {latest_version}")


@app.command()
def upgrade(
    ctx: typer.Context,
    version: Annotated[str, typer.Argument(help="New version the base package")],
):
    """Upgrade a base dependency to a newer version."""
    settings: t.RepositorySettings = ctx.obj.settings
    if not settings.managed_by_uv:
        typer.echo("Only available for installations managed by uv.")
        raise typer.Exit(1)
    package_name: str = settings.backend.base_package
    logger.info(f"Getting {package_name} constraints for version {version}")
    pyproject_path = utils.get_pyproject(settings)
    if pyproject_path and (pyproject := dependencies.parse_pyproject(pyproject_path)):
        existing_pins = dependencies.get_all_pinned_dependencies(pyproject)
        constraints = dependencies.get_package_constraints(
            package_name, version, existing_pins
        )
        logger.info(f"Updating {pyproject_path} dependencies and constraints")
        dependencies.update_pyproject(
            pyproject_path, package_name, version, constraints
        )
        # Update versions.txt
        backend_path = settings.backend.path
        version_file = (backend_path / "version.txt").resolve()
        version_file.write_text(f"{version}\n")
        logger.info("Done")
    else:
        logger.info("No pyproject.toml found")



from repoplone import _types as t
from repoplone.utils import _git as gitutils
from repoplone.utils import _github as ghutils
from repoplone.utils import changelog as chgutils
from repoplone.utils import display as dutils
from repoplone.utils import release as utils
from repoplone.utils import versions as vutils
from typing import Annotated

import typer


app = typer.Typer()


def _check_for_confirmation(
    question: str = "[bold yellow]Continue?[/bold yellow]",
    goodbye: str = "Exiting now",
) -> bool:
    status: bool = dutils.confirm(question)
    if not (status):
        dutils.print(goodbye)
        raise typer.Exit(0)
    return status


def _get_next_version(
    settings: t.RepositorySettings, original_version: str, desired_version: str
) -> tuple[str, str]:
    next_version = ""
    error = ""
    try:
        next_version = vutils.next_version(desired_version, original_version)
    except ValueError:
        next_version = ""
        error = "Invalid version."
    else:
        if not utils.valid_next_version(settings, next_version):
            error = f"The version {next_version} already exists as a tag in Git"
            next_version = ""
    return next_version, error


def _step_confirm_version(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
) -> bool:
    dutils.indented_print(f"- Bump version from {settings.version} to {next_version}")
    return _check_for_confirmation()


def _step_goodbye(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
) -> bool:
    dutils.indented_print(f"- Completed the release of version {next_version}")
    return _check_for_confirmation()


def _step_prepare_changelog(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
) -> bool:
    # Changelog
    ## First display the changelog
    new_entries, _ = chgutils.update_changelog(
        settings, draft=True, version=next_version
    )
    settings._tmp_changelog = new_entries
    text = f"{'=' * 50}\n{new_entries}\n{'=' * 50}"
    dutils.indented_print(text)
    return _check_for_confirmation()


def _step_update_repository(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
):
    if not dry_run:
        chgutils.update_changelog(settings, draft=dry_run, version=next_version)
        dutils.indented_print(f"- Updated {settings.changelogs.root} file")
    # Update next_version on version.txt
    version_file = settings.version_path
    version_file.write_text(f"{next_version}\n")
    dutils.indented_print(f"- Updated {version_file} file")
    # Update docker-compose.yml
    compose_file = settings.compose_path
    if compose_file.exists() and compose_file.is_file():
        contents = compose_file.read_text().replace(original_version, next_version)
        compose_file.write_text(contents)
        dutils.indented_print(f"- Updated {compose_file} file")
    else:
        dutils.indented_print("- No docker-compose.yml file to update")


def _step_release_backend(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
):
    utils.release_backend(settings, next_version, dry_run)
    dutils.indented_print(f"- Released {settings.backend.name}: {next_version}")


def _step_release_frontend(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
):
    next_version = vutils.convert_python_node_version(next_version)
    utils.release_frontend(settings, next_version, dry_run)
    dutils.indented_print(f"- Released {settings.frontend.name}: {next_version}")


def _step_update_git(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
):
    if not dry_run:
        repo = gitutils.repo_for_project(settings.root_path)
        gitutils.finish_release(repo, next_version)
        dutils.indented_print(f"- Created tag {next_version}")
    else:
        dutils.indented_print(f"- Skipped creating tag {next_version}")


def _step_gh_release(
    step_id: int,
    title: str,
    settings: t.RepositorySettings,
    original_version: str,
    next_version: str,
    dry_run: bool,
):
    if dry_run:
        dutils.indented_print("- Skipping GitHub release creation")
        return
    if ghutils.check_token(settings):
        msg = ghutils.create_release(settings, original_version, next_version)
        dutils.indented_print(msg)
    else:
        dutils.indented_print(
            "- Skipping GitHub release creation as you do not have a GITHUB_TOKEN"
            "environment variable set"
        )


@app.command()
def do(
    ctx: typer.Context,
    desired_version: Annotated[
        str,
        typer.Argument(
            help=(
                "Next version. Could be the version number, or "
                "a segment like: a, minor, major, rc"
            )
        ),
    ],
    dry_run: Annotated[bool, typer.Option(help="Is this a dry run?")] = False,
):
    """Release the packages in this mono repo."""
    settings: t.RepositorySettings = ctx.obj.settings
    original_version = settings.version

    dutils.print(f"\n[bold green]Release {settings.name}[/bold green]\n")
    next_version, error = _get_next_version(settings, original_version, desired_version)
    if error:
        dutils.print(error)
        typer.Exit(0)
        return

    steps = [
        ("Next version", _step_confirm_version),
        ("Display Changelog", _step_prepare_changelog),
        ("Update repository components", _step_update_repository),
        ("Release backend", _step_release_backend),
        ("Release frontend", _step_release_frontend),
        ("Commit changes, create tag", _step_update_git),
        ("Create GitHub release", _step_gh_release),
        ("Goodbye", _step_goodbye),
    ]
    total_steps = len(steps)
    for step_id, (title, func) in enumerate(steps, start=1):
        dutils.print(
            f"\n[bold green]{step_id}/{total_steps}[/bold green] [bold]{title}[/bold]"
        )
        func(step_id, title, settings, original_version, next_version, dry_run)
    raise typer.Exit(0)


@app.command()
def changelog(
    ctx: typer.Context,
):
    """Generate a draft of the final changelog."""
    settings: t.RepositorySettings = ctx.obj.settings
    original_version = settings.version
    # Changelog
    new_entries, _ = chgutils.update_changelog(
        settings, draft=True, version=original_version
    )
    dutils.print(f"{'=' * 50}\n{new_entries}\n{'=' * 50}")



import logging


__version__ = "1.0.0a1"


PACKAGE_NAME = "repoplone"


def _setup_logging():
    logger = logging.getLogger(PACKAGE_NAME)
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.INFO)
    return logger


logger = _setup_logging()



from click.testing import CliRunner as __CliRunner
from hatch import cli


class CliRunner(__CliRunner):
    def __init__(self, command):
        super().__init__()
        self._command = command

    def __call__(self, *args, **kwargs):
        # Exceptions should always be handled
        kwargs.setdefault("catch_exceptions", False)

        return self.invoke(self._command, args, **kwargs)


def get_hatch() -> CliRunner:
    return CliRunner(cli.hatch)



from rich import print  # noQA: A004
from rich.prompt import Confirm
from rich.table import Table

import textwrap


__all__ = ["print", "table"]


def table(title: str, columns: list[dict], rows: list) -> Table:
    table = Table(title=title)
    for column in columns:
        table.add_column(**column)
    for row in rows:
        table.add_row(*row)
    return table


def indented_print(text: str, prefix: str = "   "):
    print(textwrap.indent(text, prefix))


def confirm(question: str, prefix: str = "   ") -> bool:
    question = textwrap.indent(question, prefix)
    return bool(Confirm.ask(question, default=True))
